--[[
    @Liukeming 2018-1-3
    引导系统状态机
]]

local M = Class()
local N = Class()
GuideData = M --引导状态机
Guide = N --引导结构体

--引导的类型
GuideType = 
{
    Force = 0, --主线强制引导
    Common = 1, --常用的非强制引导
    Empty = 2, --无任何效果的过渡引导
}
--引导的状态
GuideStatus = 
{
    Enable = 0, --等待激活状态
    Active = 1, --已经激活（显示遮罩）
    Delay = 2, --等待延迟计时中
    Dialog = 3, --等待对话框完成（显示对话框）
    Click = 4, --等待用户点击（提升按钮，显示遮罩，手指，提示文本）
    Check = 5, --等待条件完成（可能是发送协议，播放动画，等等）
    Operation = 6, --执行完成后的操作
    Complete = -1, --完成
}

--[[
    引导结构体一开始就被加载，存在于引导状态机中
    isActive() 开启条件满足被激活 onOpen
    isFinish() 关闭条件满足被关闭 onClose
    本身有三个子状态GuideStatus

    --数据定义
    sid SID
    type GuideType引导类型
    status GuideStatus引导状态
    last 前置引导数组
    next 后置引导数组，一次判断是否可用，可用则进入，如果全部都完成了，直接报错
    save 关键帧上传到服务器要保存完成的帧数组
    alpha 半透明遮罩透明度（0-255）
    window 所属窗口名称
    dialog 剧情对话ID
    delay 延迟播放时间，毫秒
    openCondition 开启条件列表，详情见EXCEL
    closeCondition 关闭条件列表，详情见EXCEL
    operation  任务完成要执行的操作，详情见EXCEL
    holes 镂空范围数组[x,y,w,h]
    targets 点击对象[path,showArrow]
    tips 提示文本数组[x,y,w,h,id]
]]
function N:ctor(sid)
    local data = t_guide[sid]
    assume(data, '引导数据不存在，sid=' .. tostring(sid))

    self.sid = sid
    self.type = data.type
    self.status = GuideStatus.Enable
    self.last = data.last
    self.next = data.next
    self.save = data.save
    self.alpha = data.alpha
    self.window = data.window
    self.dialog = data.dialog
    self.delay = data.delay / 1000

    if self.window ~= '' then
        assert(_UIConfig[self.window], '引导触发的窗口不存在，uiName=' .. tostring(self.window))
    end

    if #self.last == 1 and self.last[1] == -1 then
        self.last = {}
    end

    if #self.next == 1 and self.next[1] == -1 then
        self.next = {}
    end
    
    if #self.save == 1 and self.save[1] == -1 then
        self.save = {}
    end

    self.openCondition = {}
    if data.openCondition and data.openCondition~=''  then
        for i,v in pairs(splitToString(data.openCondition,';')) do
            local condition = {}
            local conditionConfig, op = splitByOp(v)
            if #conditionConfig==2 then
                condition.name = conditionConfig[1]
                condition.id = Command:nameToID(condition.name)
                condition.args = splitToString(op .. ',' .. conditionConfig[2], ',')
                table.insert(self.openCondition, condition)
            end
        end
    end

    self.closeCondition = {}
    if data.closeCondition and data.closeCondition~='' then
        for i,v in pairs(splitToString(data.closeCondition,';')) do
            local condition = {}
            local conditionConfig, op = splitByOp(v)
            if #conditionConfig==2 then
                condition.name = conditionConfig[1]
                condition.id = Command:nameToID(condition.name)
                condition.args = splitToString(op .. ',' .. conditionConfig[2], ',')
                table.insert(self.closeCondition, condition)
            end
        end
    end

    self.operation = {}
    if data.operation and data.operation~='' then
        for i,v in pairs(splitToString(data.operation,';')) do
            local condition = {}
            local conditionConfig, op = splitByOp(v)
            if #conditionConfig==2 then
                condition.name = conditionConfig[1]
                condition.id = Command:nameToID(condition.name)
                --moveto 需要额外放目标窗口进去
                if condition.name == 'moveto' then
                    condition.args = splitToString(op .. ',' .. self.window .. ',' .. conditionConfig[2], ',')
                else
                    condition.args = splitToString(op .. ',' .. conditionConfig[2], ',')
                end
                table.insert(self.operation, condition)
            end
        end
    end

    self.holes = {}
    if data.holes and data.holes~='' then
        for i,v in pairs(splitToString(data.holes, ';')) do
            local holeConfig = splitToNumber(v,',')
            if #holeConfig == 5 then
                local hole =
                {
                    x = holeConfig[1],
                    y = holeConfig[2],
                    w = holeConfig[3],
                    h = holeConfig[4],
                    borderid = holeConfig[5],
                }
                table.insert(self.holes, hole)
            end
        end
    end

    self.targets = {}
    if data.targets then
        for i,v in pairs(splitToString(data.targets, ';')) do
            local targetsConfig = splitToString(v, '=')
            if #targetsConfig ==2 then
                local target = 
                {
                    path = targetsConfig[1],
                    showArrow = tonumber(targetsConfig[2])==1 and true or false,
                    button = nil,
                    parent = nil
                }
                table.insert(self.targets, target)
            end
        end
    end

    if #data.tips == 1 and data.tips[1] == -1 then
        self.tips = nil
    else
        self.tips = 
        {
            x = data.tips[1],
            y = data.tips[2],
            w = data.tips[3],
            h = data.tips[4],
            langid = data.tips[5],
        }
    end

    self.fullDialog = #self.holes == 0
end

function N:isActive()
    return self.status == GuideStatus.Active
end

function N:isFinish()
    return self.status == GuideStatus.Complete
end
--修改状态
function N:onStatus(status)
    
    if status == GuideStatus.Active then 
        print(gameDataManager.guideData.frames, '引导'..self.sid..'开启')
    elseif status == GuideStatus.Delay then 
        --print(gameDataManager.guideData.frames, '引导'..self.sid..'等待 ' .. self.delay)
    elseif status == GuideStatus.Dialog then
        --print(gameDataManager.guideData.frames, '引导'..self.sid..'对话 ' .. self.dialog)
    elseif status == GuideStatus.Click then 
        --print(gameDataManager.guideData.frames, '引导'..self.sid..'点击 是否等待：' .. tostring(self.waitingClick),
    elseif status == GuideStatus.Check then
        --print(gameDataManager.guideData.frames, '引导'..self.sid..'条件')
    elseif status == GuideStatus.Operation then
        --print(gameDataManager.guideData.frames, '引导'..self.sid..'操作')
    elseif status == GuideStatus.Complete then
        print(gameDataManager.guideData.frames, '引导'..self.sid..'完成')
    end

    self.status = status
    uiMgr:getUI(_UIConfig.UIGuide):onStatus(status)
end
--更新状态，连续更改状态时会连续校验，直到状态进入等待或完成状态
function N:update()
    local continue = true
    while(continue)
    do
        continue = self:check()
    end
end
--校验状态，有改变时返回true
function N:check()
    --状态：等待激活
    if self.status == GuideStatus.Enable then 
        return self:checkEnable()
    elseif self.status == GuideStatus.Active then 
        return self:checkActive()
    elseif self.status == GuideStatus.Delay then 
        return self:checkDelay()
    elseif self.status == GuideStatus.Dialog then
        return self:checkDialog()
    elseif self.status == GuideStatus.Click then
        return self:checkClick()
    elseif self.status == GuideStatus.Check then
        return self:checkCheck()
    elseif self.status == GuideStatus.Operation then
        return self:checkOperation()
    elseif self.status == GuideStatus.Complete then 
        return false
    else
        assert(false, '引导状态异常，status=' .. tostring(self.status))
    end
end

function N:checkEnable() 
    --前置检测
    for i,v in ipairs(self.last) do
        local guide = gameDataManager.guideData.guides[v]
        if not guide or not guide:isFinish() then
            return false
        end
    end
    --条件检测
    for i,v in ipairs(self.openCondition) do
        if not Command:check(v.id, v.args) then
            return false
        end
    end
    self:showWindow()
    self:onStatus(GuideStatus.Active)
    return true
end

function N:checkActive() 
    if self.loading==true then
        return false
    end
    if self.delay > 0 then
        self.delayTime = os.clock()
    else
        self.delayTime = 0
    end
    self:onStatus(GuideStatus.Delay)
    return true
end

function N:checkDelay() 
    --延迟时间不显示遮罩
    if self.delay > 0 and os.clock() - self.delayTime  < self.delay then
        return false
    end
    
    uiMgr:getUI(_UIConfig.UIGuide).layer2:SetActive(true)
    
    --延迟完毕，有对话框马上显示对话框
    if self.dialog>=0 then
        if not self.dialogInstance then
            self.dialogOpen = true
            local guide = self
            self.dialogInstance = uiMgr:openUI(_UIConfig.UIJuqing, true, self.dialog, function()
                guide.dialogOpen = false
            end, self.fullDialog, false, false)
        end
    else
        self.dialogOpen = false
    end
    
    --延迟完毕，等待窗口时显示遮罩
    if self.type ~= GuideType.Empty then
        local windows = uiMgr:isUIConfigExistOpenList(_UIConfig[self.window])
        if #windows == 0 or windows[1][1]._started == false then
            print('waiting guide target window delay for', self.window)
            return false
        end
    end

    self:onStatus(GuideStatus.Dialog)
    return true
end

--[[
    fullDialog 全屏剧情，等待剧情结束
    否则是点击剧情
]]
function N:checkDialog() 
    if self.fullDialog==true then
        if self.dialogOpen==true then
            return false
        else
            self.waitingClick = false
        end
    else
        self.waitingClick = true
    end

    self.dialogInstance = nil
    self:onStatus(GuideStatus.Click)
    return true
end

function N:checkClick() 
    if self.waitingClick == true then
        return false
    else
        --非全屏要手动关闭对话框
        if self.dialog>=0 and self.fullDialog==false then
            uiMgr:closeUIConfig(_UIConfig.UIJuqing)
        end
        self:onStatus(GuideStatus.Check)
        return true
    end
end

function N:checkCheck()
    --条件检测
    for i,v in ipairs(self.closeCondition) do
        if not Command:check(v.id, v.args) then
            return false
        end
    end

    --上传点
    if self.save and #self.save>0 then
        gameDataManager.guideData:saveStatus(self.save)
    end

    self:onStatus(GuideStatus.Operation)
    return true
end

function N:checkOperation()
    --结束完成的操作
    for i,v in ipairs(self.operation) do
        Command:execute(v.id, v.args)
    end

    --通过
    self:onStatus(GuideStatus.Complete)
    self:hideWindow()
    return true
end

--打开或者重新打开窗口
function N:showWindow()
    if self.type == GuideType.Empty then
        self.ui = nil
        self.loading = false
        return
    else
        --[[
        self.loading = true
        local windows = uiMgr:isUIConfigExistOpenList(_UIConfig.UIGuide)
        if #windows > 0 then
            self.ui = windows[1][1]
            self.ui:show(self)
        else
            self.ui = uiMgr:openUI(_UIConfig.UIGuide, true,  self)
        end
        ]]

        self.loading = false --预加载，不需要再等待了
        uiMgr:getUI(_UIConfig.UIGuide):show(self)
    end
end

--隐藏窗口
function N:hideWindow()
    --[[
    if self.ui then
        self.ui:hide()
        self.ui = nil
    end
    ]]
    uiMgr:getUI(_UIConfig.UIGuide):hide()
end

--[[
    引导状态机负责某个引导的执行以及多个引导的关联
    除了初始化的引导字典
    还额外生成了基于窗口的可用(Enable)引导的索引字典用于方便在激活时查找和条件判断
]]
function M:ctor()
    self.frames = 1
    self.guides = {}
    self.windows = {}
    for i,v in pairs(t_guide) do
        local guide = Guide.new(v.sid)
        self.guides[v.sid] = guide

        if not self.windows[guide.window] then
            self.windows[guide.window] = {}
        end

        self.windows[guide.window][v.sid] = guide
    end
    self.guide = nil
end

function M:init(data)
    self.frames = 1
    self.guides = {}
    self.windows = {}
    for i,v in pairs(t_guide) do
        local guide = Guide.new(v.sid)
        self.guides[v.sid] = guide

        if not self.windows[guide.window] then
            self.windows[guide.window] = {}
        end

        self.windows[guide.window][v.sid] = guide
    end
    self.guide = nil
end

function M:loadStatus(sids)
    --重载，防止切换账号造成状态异常问题
    self:init()

    if not sids or #sids==0 then
        return
    end
    local str = ''
    for i,v in ipairs(sids) do
        if self.guides[v.key] then
            self.guides[v.key].status = GuideStatus.Complete
            str = str .. tostring(v.key) .. ','
        end
    end
    print('加载已完成引导=' .. str)
end

function M:saveStatus(sids)
    local kvs = {}
    local str = ''
    for i,v in ipairs(sids) do
            table.insert(kvs, v)
            table.insert(kvs, 1)
            str = str .. tostring(v) .. ','
    end
    print('上传已完成引导=' .. str)
    _gameTools.sendMsg(cs_playerSET_CLIENT_STATES, {6, kvs}, function() end)
end

function M:pause()
    self.paused = true
end

function M:resume()
    self.paused = nil
    self:update()
end

function M:update(noAddFrame)
    if not noAddFrame then
        self.frames = self.frames + 1
    end

    if self.paused == true then
        return
    end
    
    if not self.guide then
        return
    else
        self.guide:update()
    end

    while(self.guide and self.guide:isFinish())
    do
        self:next()
    end
end

function M:next()
    if #self.guide.next == 0 or self.guide.next[1]==-1 then
        self.guide = nil
        --连续进行的引导不会关闭窗口，所有引导都结束后窗口关闭，遮罩关闭
        --if uiMgr:isUIConfigExistOpenList(_UIConfig.UIGuide) then
        --    uiMgr:closeUIConfig(_UIConfig.UIGuide)
        --end
        uiMgr:getUI(_UIConfig.UIGuide):hide()

        --继续激活下一帧
        self:checkActivable()
    else
        --下一帧的引导ID数组里寻找可以激活的引导
        local nextid = nil
        for i,v in pairs(self.guide.next) do
            if nextid == nil and self.guides[v].status == GuideStatus.Enable then
                nextid = v
            end
        end
        assert(nextid ~= nil, '配置错误，不允许出现下一帧的所有可能引导都已完成的情况，当前引导sid=' .. tostring(self.guide.sid))
        self.guide = self.guides[nextid]
        self.guide:update()
    end
end

--寻找当前窗口可激活的引导
function M:checkActivable()
    --不管激活与否，都刷新未解锁功能
    uiMgr:getUI(_UIConfig.UIMainCity):refreshTitleStatus()

    if self.paused == true then
        return
    end
    
    if gameDef.GuideEnable==false then
        return
    end

    if self.guide then
        return
    end

    local topWindow = uiMgr:getTopUI()

    if not topWindow or topWindow._started==false then
        return
    end

    --新角色剧情没触发则不刷新新手引导
    if not gameDataManager.isLoggedIn or gameDataManager.getPlayerClientStateByKey(_EClientStates.KaiChangDongHua) == nil then
        return
    end

    local guides = self.windows[topWindow.uiName]

    if not guides then
        return
    end

    for i,v in pairs(guides) do
        if v.status == GuideStatus.Enable and v:check() and v:isActive() then
            self.guide = v
            self:update(true)
            return
        end
    end
end

--捕捉到一个点击事件
function M:onClick(go)
    if self.guide and self.guide.status==GuideStatus.Click then
        uiMgr:getUI(_UIConfig.UIGuide):onClickButton(go)
    end
end

gameDataManager.guideData = GuideData.new()